%%
%%  JFKi
%%  
%%  Aiello et.all., Efficient, DoS-Resistant, Secure Key Exchange for 
%%  Internet Protocols
%%


% Functors:

    % cryptographic primitives
    fun sig/2.  % sig(K,M) -- a non-message-recovering signature on M
    fun enc/2.  % enc(K,M) -- encryption of M with key K
    fun h/2.    % h(K,M) -- keyed hash (HMAC) of message M using key K

    % participant names: a,b are honest, c is dishonest
    fun a/0.
    fun b/0.
    fun c/0.

    % CA
    fun ca/2.

    % constants
    fun g/0.
    fun sec/0.
    fun n0/0.
    fun n1/0.
    fun n2/0.

    % nonces and keys
    fun i/1.    % i(X) -- a nonce (exponent) generated by X as an initiator 
    fun r/1.    % r(X) -- a nonce (exponent) generated by X as a responder 
    fun ni/2.   % ni(X,Y) -- a nonce of X as an initiator generated for communication with Y
    fun nr/2.   % ni(X,Y) -- a nonce of Y as a responder generated for communication with Y
    fun hk/1.   % hk(X) -- a transient hash key private to the responder X.

    fun k/1.    % k(K) -- private key of X.
    fun pub/1.

    fun completedI/3.
    fun completedR/3.


%%%%% AUXILIARY PREDICATES %%%%%%%%%%%%

participant(a).
% participant(b).
participant(c).
honest(a).

%%
% The following predicates specify which session will be
% taken into consideration.

% roleI(X,Y)  ---  X as an initiator runs the protocol with Y as responder 
% roleR(X,Y)  ---  X as a responder runs the protocol with Y as initiator 
%

roleI(a, a).
% roleI(a, b).
roleI(a, c).
% roleI(b, a). 
% roleI(b, c). 

roleR(a, a). 
% roleR(a, b). 
roleR(a, c). 
% roleR(b, a).
% roleR(b, c).

kir(K, GIR, NI, NR) :-   K = h(GIR, (NI,NR,n0)).
ke(K, GIR, NI, NR)  :-   K = h(GIR, (NI,NR,n1)).
ka(K, GIR, NI, NR)  :-   K = h(GIR, (NI,NR,n2)).

macenc(ME, Who, KE,KA, X) :-
    ME = h(KA, (enc(KE,X),Who)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The secret:

    query sec.

% Intruder rules

    rule:: sig(_,M) -> M. 
    rule:: sig(K,_) -> pub(K). 
    rule:: K, M -> sig(K,M).
    rule:: K, M -> enc(K,M).
    rule:: K, enc(K,M) -> M.
    rule:: K, M -> h(K,M).
    rule:: h(_,M) -> M.

% Modelling CA

    % Honest participants have regitered their public keys properly:
    rule:: ca(A,pub(k(A)))  :- honest(A).

    % A dishonest participant can register anything as his public key:
    rule:: X -> ca(c,X).

% Initial intruder knowledge

    rule:: g.
    rule:: n0.
    rule:: n1.
    rule:: n2.
    rule:: i(c).
    rule:: r(c).
    rule:: X              :- participant(X).
    rule:: ni(c,X)        :- participant(X).
    rule:: nr(c,X)        :- participant(X).
    rule:: hk(c).
    rule:: pub(k(X))      :- participant(X).
    rule:: k(c).

% Protocol Rules (the main session):

    %%
    % (1) [I->R]
    %
    % We use R (the name of the responder) instead of ID_{R'}.
    %
    rule::  ni(I,R),  g^i(I), R  
        :- roleI(I,R).


    %%
    % (2) [R->I]
    %
    % We ommit grpinfo_R and IP_I.
    %
    rule::  NI, _GI, R ->  
                NI, nr(R,I), g^r(R), R,
                sig(k(R), g^r(R)),
                h(hk(R), (g^r(R), nr(R,I), NI))
    :- roleR(R,I).


    %%
    % (3) [I->R]
    %
    % We ommit sa.
    %
    rule::  ni(I,R), NR, GR, R, ca(R,pub(KR)), sig(KR,GR), AuthR ->
                ni(I,R), NR, g^i(I), GR,
                AuthR,
                MEI
    :- roleI(I,R), 
       ke(KE, GR^i(I), ni(I,R), NR),
       ka(KA, GR^i(I), ni(I,R), NR),
       macenc(MEI, I, KE, KA, (I, sig(k(I), (ni(I,R),NR,g^i(I),GR,R)))).

    %%
    % (4) [R->I]
    %
    % We ommit sa and sa'.
    %
    rule::
        NI, ca(I,pub(KI)), nr(R,I), GI, g^r(R), h(hk(R), (g^r(R), nr(R,I), NI)), MEI ->
            MER, completedR(R, I, GI^r(R))
    :- roleR(R,I),
       ke(KE, GI^r(R), NI, nr(R,I)),
       ka(KA, GI^r(R), NI, nr(R,I)),
       % kir(KIR, GI^r(R), NI, nr(R,I)),
       macenc(MEI, I, KE, KA, (I, sig(KI, (NI,nr(R,I),GI,g^r(R),R)))),
       macenc(MER, R, KE, KA, sig(k(R), (NI,nr(R,I),GI,g^r(R),I))).

    %%
    % I comleted:
    %
    rule:: ca(R,pub(KR)), MER  ->  completedI(I, R, GR^i(I))
    :-
       roleI(I,R),
       ke(KE, GR^i(I), ni(I,R), NR),
       ka(KA, GR^i(I), ni(I,R), NR),
       % kir(KIR, GR^i(I), ni(I,R), NR),
       macenc(MER, R, KE, KA, sig(KR, (ni(I,R),NR,g^i(I),GR,I))).


%%%%%%

    %%%
    %% should not fire
    %
    
    rule::  K, completedI(a,a,K) -> sec.  

    rule:: completedI(a,a,K), completedR(a,c,K) -> sec.
    rule:: completedI(a,c,K), completedR(a,a,K) -> sec.
    
    rule:: g^i(a)^r(a) -> sec.
   
    %%%
    %% should fire
    %
    % rule::  completedI(a,a,K), completedR(a,a,K) -> sec.  
    % rule::  K, completedI(a,c,K) -> sec.  
    % rule::  K, completedR(a,c,K) -> sec.  
    
